(* TP2: Source !\ahrefurl{string.ml} *)

(* valuateur, appel par nom/par valeur, rduction  forte  *)


open S.Ast (* Documentation de !\pcfmodule{S.Ast}! *)


(* Substitition complte : attention aux captures.
   Rsolu par changement systmatique des lieurs
   traverss *)

exception Error of string

module VS = Var.Vars (* Ensemble de variables *)

(* Renvoie une variable qui n'appartient pas  l'ensemble xs *)

let not_in xs =
  let rec do_rec k =
    let v = Printf.sprintf "_x%02i" k in (* ou "_x" ^ string_of_int k *)
    if VS.mem v xs then
      do_rec (k+1)
    else
      v in
  do_rec 0


(* Calcul des  variables libres *)
let rec free_vars e = match e with
| Num _ -> VS.empty
| Var x -> VS.singleton x
(* Noter le motif-ou *)
| Op (_,e1,e2) | App (e1,e2) ->
    VS.union (free_vars e1) (free_vars e2)
| Ifz (e1,e2,e3) ->
    VS.union (free_vars e1)
      (VS.union (free_vars e2) (free_vars e3))
(* Encore un motif-ou *)
| Fun (x,e1) | Fix (x,e1) ->
    VS.remove x (free_vars e1)
| Let (x,e1,e2) ->
    VS.union
      (free_vars e1)
      (VS.remove x (free_vars e2))

(* Substitution dans le cas gnral *)  
let rec  subst e1 x e2 = match e1 with
| Num _ -> e1
| Var y -> if x = y then e2 else e1
| Op (op,a1,a2) -> Op (op, subst a1 x e2, subst a2 x e2)
| App (a1,a2) -> App (subst a1 x e2, subst a2 x e2)
| Ifz (a1,a2,a3) -> Ifz (subst a1 x e2, subst a2 x e2, subst a3 x e2)
| Fun (y,a1) ->
    let y,a1 = subst_binder y a1 x e2 in
    Fun (y,a1)
| Fix (y,a1) ->
    let y,a1 = subst_binder y a1 x e2 in
    Fix (y,a1)
| Let (y,a1,a2) ->
    let y,a2 = subst_binder y a2 x e2 in
    Let (y,subst a1 x e2,a2)

and subst_binder y e1 x e2 =
  let free_2 = free_vars e2 in
  if x = y then
    y,e1
  else if VS.mem y free_2 then
    (* Ici il y a capture : une variable de l'argument est gale  x *)
    let z = not_in free_2 in
    z,subst (alpha_conv e1 y z) x e2
  else
    y,subst e1 x e2

and alpha_conv e x y = subst e x (Var y)
    

(*****************)
(* Appel par nom *)
(*****************)


(* valuation en forme normale de tte [failble] *)
let rec weakn e = match e with
(* Ce sont les valeurs *)
| Num _|Fun (_,_) | Var _ -> e
(* Appel  une function auxilaire (partage de code) *)
| Op (op,e1,e2) ->
    let w1 = weakn e1
    and w2 = weakn e2 in
    begin match w1,w2 with
    | Num n1, Num n2 -> Num (Op.to_fun op n1 n2)
    | _,_ -> Op (op,w1,w2)
    end
(* Noter l'emploi de  if  dans l'argument de weakn ci-dessous *)
| Ifz (e1,e2,e3) ->
    let w1 = weakn e1 in
    begin match w1 with
    | Num 0 -> weakn e2
    | Num _ -> weakn e3
    | _ -> Ifz (w1, e2, e3)
    end
| App (e1,e2) ->
    let w1 =  weakn e1 in
    begin match w1 with
    | Fun (x,body) -> weakn (subst body x e2)
    | _ -> App (w1,e2)
    end
| Let (x,ex,e) -> weakn (subst e x ex)
(* valuer  gauche d'abord *)
| Fix (x,ex) -> weakn (subst ex x e)

let rec kont w = match w with
  | Num _|Var _ -> w
  | Op (op,w1,w2) -> Op (op,kont w1, kont w2)
  | Ifz (w1,e2,e3) -> Ifz (kont w1,eval e2,eval e3)
  | App (w1,e2) -> App (kont w1,eval e2)
  | Fun (x,e) -> Fun (x,eval e)
  | Let _|Fix _ -> assert false

and eval e = kont (weakn e)

let evaln = eval


(********************)
(* Appel par valeur *)
(********************)

let rec evalv _e = raise (Error "Not implemented yet")
